@@ -15,7 +15,7 @@
 	  (restore-history #f)      ; function to restore history buffer entries from a file
 	  (helpers ())              ; list of functions displaying help strings 
 	  (run #f)                  ; function that fires up a REPL
-	  (top-level-let (sublet (rootlet) :** #f)) ; environment in which evaluation takes place
+	  (top-level-let (sublet (rootlet))) ; environment in which evaluation takes place
 	  (repl-let                 ; environment for keymap functions to access all the REPL innards (cursor-position etc)
 	   
       (with-let (sublet *libc*)
@@ -28,13 +28,20 @@
 	    (call-with-exit
 	     (lambda (return)
 	       (for-each
-		(lambda (symbol)
-		  (let* ((sym (symbol->string symbol))
+		(lambda (par)
+		  (let* ((sym (symbol->string par))
 			 (sym-len (length sym)))
 		    (when (and (>= sym-len text-len)
 			       (string=? text (substring sym 0 text-len)))
 		      (if match
-			  (return text)
+			  ;; more than one match, save the longest text that all syms match
+			  (do ((min-len (min (string-length match) sym-len))
+			       (i text-len (+ i 1)))
+			      ((or (= i min-len)
+				   (not (char=? (match i) (sym i))))
+			       (if (= min-len text-len)
+				   (return text)
+				   (set! match (substring match 0 i)))))
 			  (set! match sym)))))
 		st)
 	       (or match text)))))
@@ -153,7 +160,7 @@
 	    (format port "~NC(set! histpos ~D)~%" spaces #\space histpos)
 	    (format port "~NC(set! histsize ~D)~%" spaces #\space histsize)
 	    (let-temporarily (((*s7* 'print-length) (* 2 histsize)))
-	      (format port "~NC(set! histbuf ~A)" spaces #\space (object->string histbuf))))
+	      (format port "~NC(set! histbuf ~A)" spaces #\space histbuf)))
 	  
 	  (define* (save-history (file "repl-history.scm"))
 	    (call-with-output-file file
@@ -187,8 +194,9 @@
 	    ;; -------- evaluation ---------
 	    (define (badexpr h)            ; *missing-close-paren-hook* function for Enter command
 	      (let ((ow (owlet)))
-		(if (ow 'error-file)
-		    (error "missing close paren in ~S" (ow 'error-file))
+		(if (and (ow 'error-file)
+			 (not (equal? (ow 'error-file) "repl.scm")))
+		    (error 'syntax-error "missing close paren in ~S" (ow 'error-file))
 		    (set! (h 'result) 'string-read-error))))
 	    
 	    (define (shell? h)             ; *unbound-variable-hook* function, also for Enter
@@ -203,11 +211,8 @@
 				(zero? (system (string-append "command -v " var-name " >/dev/null"))))
 		       (set! unbound-case #t)
 		       (if (procedure? ((rootlet) 'system))
-			   (begin
-			     (set! ((*repl* 'top-level-let) '**) (((rootlet) 'system) cur-line #t))
-			     (display ((*repl* 'top-level-let) '**) *stderr*))
-			   (set! ((*repl* 'top-level-let) '**) (system cur-line)))
-		       (set! (h 'result) (symbol " ")))))))
+			   (set! (h 'result) (((rootlet) 'system) cur-line #f))
+			   (set! (h 'result) #f)))))))
 	    
 	    (define (with-repl-let body)
 	      ;; for multiline edits, we will use *missing-close-paren-hook* rather than try to parse the input ourselves.
@@ -237,22 +242,22 @@
 			 (set! (hook-functions *unbound-variable-hook*) old-unbound-var-hook))))
 		
 		  (let ((new-load 
-			 (let ((documentation "this is the repl's load replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values string? let?)))
+			 (let ((+documentation+ "this is the repl's load replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values string? let?)))
 			   (lambda* (file (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (load file e)) repl-hooks))))
+			     (dynamic-wind repl-hooks (lambda () (load file e)) original-hooks))))
 			
 			(new-eval 
-			 (let ((documentation "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values list? let?)))
+			 (let ((+documentation+ "this is the repl's eval replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values list? let?)))
 			   (lambda* (form (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (eval form e)) repl-hooks))))
+			     (dynamic-wind repl-hooks (lambda () (eval form e)) original-hooks))))
 			
 			(new-eval-string 
-			 (let ((documentation "this is the repl's eval-string replacement; its default is to use the repl's top-level-let.")
-			       (signature '(values string? let?)))
+			 (let ((+documentation+ "this is the repl's eval-string replacement; its default is to use the repl's top-level-let.")
+			       (+signature+ '(values string? let?)))
 			   (lambda* (str (e (*repl* 'top-level-let)))
-			     (dynamic-wind original-hooks (lambda () (eval-string str e)) repl-hooks)))))
+			     (dynamic-wind repl-hooks (lambda () (eval-string str e)) original-hooks)))))
 		    
 		    (dynamic-wind
 			(lambda ()
@@ -269,61 +274,61 @@
 	    
 	    
 	    ;; -------- match parens --------
-	    (define (check-parens)
-
-	      (define (char-constant? pos)
-		(and (> pos 2)
-		     (char=? (cur-line (- pos 1)) #\\)
-		     (char=? (cur-line (- pos 2)) #\#)))
-	    
-	      (let ((endpos (- cursor-pos 1)))
-		(if (or (<= cursor-pos 1)
-			(not (char=? (cur-line endpos) #\)))      ; ")" on left of cursor
-			(char-constant? endpos))                  ; it's not "#\)"
-		    (if (number? red-par-pos)
-			(set! red-par-pos #f))
-		    (let ((oparens ())
-			  (new-red-pos #f))
-		      (do ((i 0 (+ i 1)))
-			  ((>= i endpos))
-			(case (cur-line i)
-			  ((#\()
-			   (set! oparens (cons i oparens)))
-			  
-			  ((#\))
-			   (if (pair? oparens)
-			       (set! oparens (cdr oparens))))
-			  
-			  ((#\;)
-			   (do ((k (+ i 1) (+ k 1)))
-			       ((or (>= k endpos)
-				    (char=? (cur-line k) #\newline))
-				(set! i k)
-				(if (>= i endpos)                   ; (f1 "(+ 1 3) should not show first paren as a match (similarly below)
-				    (set! oparens ())))))
-			  
-			  ((#\")
-			   (do ((k (+ i 1) (+ k 1)))
-			       ((or (>= k endpos)
-				    (and (char=? (cur-line k) #\")
-					 (not (char=? (cur-line (- k 1)) #\\))))
-				(set! i k)
-				(if (>= i endpos)
-				    (set! oparens ())))))
-			  
-			  ((#\#)
-			   (if (char=? (cur-line (+ i 1)) #\|)
+	    (define check-parens
+	      (let ((char-constant? 
+		     (lambda (pos)
+		       (and (> pos 2)
+			    (char=? (cur-line (- pos 1)) #\\)
+			    (char=? (cur-line (- pos 2)) #\#)))))
+		(lambda ()
+		  (let ((endpos (- cursor-pos 1)))
+		    (if (or (<= cursor-pos 1)
+			    (not (char=? (cur-line endpos) #\)))      ; ")" on left of cursor
+			    (char-constant? endpos))                  ; it's not "#\)"
+			(if (number? red-par-pos)
+			    (set! red-par-pos #f))
+			(let ((oparens ())
+			      (new-red-pos #f))
+			  (do ((i 0 (+ i 1)))
+			      ((>= i endpos))
+			    (case (cur-line i)
+			      ((#\()
+			       (set! oparens (cons i oparens)))
+			      
+			      ((#\))
+			       (if (pair? oparens)
+				   (set! oparens (cdr oparens))))
+			      
+			      ((#\;)
 			       (do ((k (+ i 1) (+ k 1)))
 				   ((or (>= k endpos)
-					(and (char=? (cur-line k) #\|)
-					     (char=? (cur-line (+ k 1)) #\#)))
-				    (set! i (+ k 1))
+					(char=? (cur-line k) #\newline))
+				    (set! i k)
+				    (if (>= i endpos)                   ; (f1 "(+ 1 3) should not show first paren as a match (similarly below)
+					(set! oparens ())))))
+			      
+			      ((#\")
+			       (do ((k (+ i 1) (+ k 1)))
+				   ((or (>= k endpos)
+					(and (char=? (cur-line k) #\")
+					     (not (char=? (cur-line (- k 1)) #\\))))
+				    (set! i k)
 				    (if (>= i endpos)
-					(set! oparens ()))))))))
-		      (if (pair? oparens)
-			  (set! new-red-pos (car oparens)))
-		      (unless (equal? new-red-pos red-par-pos)
-			(set! red-par-pos (and (number? new-red-pos) new-red-pos)))))))
+					(set! oparens ())))))
+			      
+			      ((#\#)
+			       (if (char=? (cur-line (+ i 1)) #\|)
+				   (do ((k (+ i 1) (+ k 1)))
+				       ((or (>= k endpos)
+					    (and (char=? (cur-line k) #\|)
+						 (char=? (cur-line (+ k 1)) #\#)))
+					(set! i (+ k 1))
+					(if (>= i endpos)
+					    (set! oparens ()))))))))
+			  (if (pair? oparens)
+			      (set! new-red-pos (car oparens)))
+			  (unless (equal? new-red-pos red-par-pos)
+			    (set! red-par-pos (and (number? new-red-pos) new-red-pos)))))))))
 	    
 	    
 	    ;; -------- indentation --------
@@ -412,10 +417,13 @@
 	    
 	    ;; to enable mouse click coords (in xterm anyway), (format *stderr* "~C[?9h" #\escape)
 	    ;;    disable: (format *stderr* "~C[?9l" #\escape)
+	    ;;    this is X10_MOUSE -- probably better to use SET_ANY_EVENT_MOUSE 1003
 	    ;; while enabled, mouse selection instead sends coords to repl (so it's annoying)
 	    ;;    also it's sticky!  exit repl does not clear this flag so mouse is effectively dead
-	    ;; upon click, we get ESC [ M bxy -- need to poll for this?
-	    
+	    ;; upon click, we get ESC [ M b x y (b = button info, x/y are coords starting at 32=0 -- 6 chars received)
+	    ;; to trap C-C (define (sigcatch no) (tty-reset) (#_exit)) (if (equal? (signal SIGINT sigcatch) SIG_ERR) (#_exit))
+	    ;; to see click, add a meta-key for M
+	    ;; all this apparently only works in xterm, so I hesitate to build on it here.
 	    
 	    ;; -------- display --------
 	    (define (display-prompt)
@@ -430,34 +438,33 @@
 	      (display-prompt)
 	      (cursor-bounds))
 	    
-	    (define (display-line start end)
-
-	      (define (bold text) 
-		(format #f "~C[1m~A~C[0m" #\escape text #\escape)) 
-
-	      ;; if a line wraps, it will confuse the redisplay/cursor positioning code. so truncate the display
-	      (let ((line-len (- (+ end prompt-length 1) start)))
-		(if (>= line-len last-col)
-		    (set! end (- (+ end line-len) last-col))))
-	      
-	      (if (and (integer? red-par-pos)
-		       (<= start red-par-pos)
-		       (< red-par-pos end))
-		  (string-append
-		   (format #f (if (zero? start)
-				  (values "~A" prompt-string)
-				  (values "~NC" prompt-length #\space)))
-		   (format #f (if (= start red-par-pos)
-				  (values "~A~A"
-					  (bold (red "(")) 
-					  (substring cur-line (+ start 1) end))
-				  (values "~A~A~A"
-					  (substring cur-line start red-par-pos) 
-					  (bold (red "(")) 
-					  (substring cur-line (+ red-par-pos 1) end)))))
-		  (format #f (if (zero? start)
-				 (values "~A~A" prompt-string (substring cur-line 0 end))
-				 (values "~NC~A" prompt-length #\space (substring cur-line start end))))))
+	    (define display-line 
+	      (let ((bold (lambda (text) 
+			    (format #f "~C[1m~A~C[0m" #\escape text #\escape))))
+		(lambda (start end)
+		  ;; if a line wraps, it will confuse the redisplay/cursor positioning code. so truncate the display
+		  (let ((line-len (- (+ end prompt-length 1) start)))
+		    (if (>= line-len last-col)
+			(set! end (- (+ start last-col) prompt-length 1))))
+		  
+		  (if (and (integer? red-par-pos)
+			   (<= start red-par-pos)
+			   (< red-par-pos end))
+		      (string-append
+		       (format #f (if (zero? start)
+				      (values "~A" prompt-string)
+				      (values "~NC" prompt-length #\space)))
+		       (format #f (if (= start red-par-pos)
+				      (values "~A~A"
+					      (bold (red "(")) 
+					      (substring cur-line (+ start 1) end))
+				      (values "~A~A~A"
+					      (substring cur-line start red-par-pos) 
+					      (bold (red "(")) 
+					      (substring cur-line (+ red-par-pos 1) end)))))
+		      (format #f (if (zero? start)
+				     (values "~A~A" prompt-string (substring cur-line 0 end))
+				     (values "~NC~A" prompt-length #\space (substring cur-line start end))))))))
 	    
 	    (define (display-cursor)
 	      (do ((row 0)
@@ -507,7 +514,7 @@
 		      ((null? lst))
 		    (let ((str ((car lst) c)))
 		      (move-cursor i col)
-		      (format *stderr* "~C[K| ~A"  #\escape (if (> (length str) col) (substring str 0 (- col 1)) str))))
+		      (format *stderr* "~C[K| ~A" #\escape (if (> (length str) col) (substring str 0 (- col 1)) str))))
 		  
 		  (move-cursor (+ 2 (length (*repl* 'helpers))) col)
 		  (format *stderr* "+~NC" (- col 2) #\-)
@@ -535,7 +542,7 @@
 			       prompt-row))
 		     (lambda (c)
 		       (format #f "c: ~S ~D, start: ~A, end: ~A" 
-			       (object->string c #t) 
+			       c
 			       (char->integer c)	
 			       (start-of-line cursor-pos)	
 			       (end-of-line cursor-pos))))))
@@ -548,29 +555,29 @@
 	    (define keymap (dilambda
 			    (lambda (c)
 			      (cond ((char? c) 
-				     (keymap-functions (char->integer c)))
+				     (vector-ref keymap-functions (char->integer c)))
 				    ((integer? c) 
-				     (keymap-functions c))
+				     (vector-ref keymap-functions c))
 				    ((not (string? c))
 				     (error 'wrong-type-arg "keymap takes a character or string argument"))
 				    ((= (length c) 1) 
-				     (keymap-functions (char->integer (c 0))))
+				     (vector-ref keymap-functions (char->integer (c 0))))
 				    ((and (= (length c) 2)
 					  (char=? (c 0) #\escape))
-				     (meta-keymap-functions (char->integer (c 1))))
+				     (vector-ref meta-keymap-functions (char->integer (c 1))))
 				    (else (lambda (c) #t))))
 			    (lambda (c f)
 			      (cond ((char? c) 
-				     (set! (keymap-functions (char->integer c)) f))
+				     (vector-set! keymap-functions (char->integer c) f))
 				    ((integer? c) 
-				     (set! (keymap-functions c) f))
+				     (vector-set! keymap-functions c f))
 				    ((not (string? c))
 				     (error 'wrong-type-arg "set! keymap takes a character or string first argument"))
 				    ((= (length c) 1)
-				     (set! (keymap-functions (char->integer (c 0))) f))
+				     (vector-set! keymap-functions (char->integer (c 0)) f))
 				    ((and (= (length c) 2)
 					  (char=? (c 0) #\escape))
-				     (set! (meta-keymap-functions (char->integer (c 1))) f))))))
+				     (vector-set! meta-keymap-functions (char->integer (c 1)) f))))))
 	    
 	    (define C-a 1)     ; #\x01 etc
 	    (define C-b 2)
@@ -617,13 +624,12 @@
 	      (set! cur-row (+ cur-row 1)))
 	    
 	    (define (word-break pos)                 ; assume we're at the start of a word
-	      (let ((len (length cur-line)))
-		(let loop ((i pos))
-		  (if (or (>= i len)
-			  (not (or (char-alphabetic? (cur-line i))
-				   (char-numeric? (cur-line i)))))
-		      i
-		      (loop (+ i 1))))))
+	      (do ((len (length cur-line))
+		   (i pos (+ i 1)))
+		  ((or (>= i len) 
+		       (not (or (char-alphabetic? (cur-line i)) 
+				(char-numeric? (cur-line i)))))
+		   i)))
 	    
 	    (define (save-line)
 	      (set! prev-line (cons (cons cursor-pos (copy cur-line)) prev-line)))
@@ -644,15 +650,15 @@
 	      
 	      (do ((i 0 (+ i 1)))
 		  ((= i 32))
-		(set! (keymap-functions i) no-op-keyfunc))
+		(vector-set! keymap-functions i no-op-keyfunc))
 	      
 	      (do ((i 32 (+ i 1)))
 		  ((= i 256))
-		(set! (keymap-functions i) main-keyfunc))
+		(vector-set! keymap-functions i main-keyfunc))
 	      
 	      (do ((i 0 (+ i 1)))
 		  ((= i 256))
-		(set! (meta-keymap-functions i) no-op-keyfunc)))
+		(vector-set! meta-keymap-functions i no-op-keyfunc)))
 	    
 	    
 	    ;; -------- cursor movement 
@@ -894,17 +900,19 @@
 				    ;; get the newline out if the expression does not involve a read error
 				    (let ((form (with-input-from-string cur-line #_read)))    ; not libc's read
 				      (newline *stderr*)
-				      (let ((val (eval form (*repl* 'top-level-let))))
+				      (let ((val ((lambda args
+						    (if (or (null? args)   ; try to trap (values) -> #<unspecified>
+							    (and (unspecified? (car args))
+								 (null? (cdr args))))
+							#<unspecified>
+							(if (pair? (cdr args))
+							    (cons 'values args)
+							    (car args))))
+						  (eval form (*repl* 'top-level-let)))))
+					(eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) ',val) (rootlet))
 					(if unbound-case
 					    (set! unbound-case #f)
-					    (begin
-					      (format *stderr* "~S~%" val)
-					      ;; this set! of '** has one odd consequence: if val is a lambda expression
-					      ;;   find_closure in s7 will fallback on the current environment trying to
-					      ;;   find an associated name, and the only thing it finds is '**!  So,
-					      ;;   when we type **, we get back **, which seems perverse.  I suppose
-					      ;;   we could trap the string above, see "**" and change to ~W or something.
-					      (set! ((*repl* 'top-level-let) '**) val))))))))
+					    (format *stderr* "~S~%" val)))))))
 			       
 			       (lambda (type info)
 				 (pop-history)               ; remove last history entry
@@ -912,13 +920,20 @@
 				 (return))))
 			   
 			   (lambda (type info)
-			     (format *stderr* "~A:" (red "error"))
-			     (if (and (pair? info)
-				      (string? (car info)))
-				 (format *stderr* " ~A" (apply format #f info))
-				 (if (not (null? info))
-				     (format *stderr* " ~A" info)))
-			     (newline *stderr*)))
+			     (with-let (unlet)
+			       (format *stderr* "~A:" (red "error"))
+			       (let ((op (*s7* 'print-length)))
+				 (if (< op 32) (set! (*s7* 'print-length) 32))
+				 (if (and (pair? info)
+					  (string? (car info)))
+				     (format *stderr* " ~A" (apply format #f info))
+				     (if (not (null? info))
+					 (format *stderr* " ~A" info)))
+				 (if (< op 32) (set! (*s7* 'print-length) op)))
+			       (with-repl-let
+				(lambda ()
+				  (eval `(define ,(string->symbol (format #f "<~D>" (+ (length histtop) 1))) 'error) (rootlet))))
+			       (newline *stderr*))))
 			 
 			 (push-line (copy cur-line))
 			 (new-prompt))))))
@@ -941,146 +956,158 @@
 	    (let ()
 	      ;; Meta key is a problem on the Mac, so I'll package these for easier disposal
 
-	      (define (fixup-new-line)
-		(define (count-newlines line)
-		  (do ((len (length line))
-		       (newlines 0)
-		       (i 0 (+ i 1)))
-		      ((= i len) newlines)
-		    (if (char=? (cur-line i) #\newline)
-			(set! newlines (+ newlines 1)))))
+	      (define fixup-new-line
+		(let ((count-newlines 
+		       (lambda (line)
+			 (do ((len (length line))
+			      (newlines 0)
+			      (i 0 (+ i 1)))
+			     ((= i len) newlines)
+			   (if (char=? (cur-line i) #\newline)
+			       (set! newlines (+ newlines 1)))))))
+		  (lambda ()
+		    (set! cursor-pos (length cur-line))
+		    (let ((newlines (count-newlines cur-line)))
+		      (when (< last-row (+ prompt-row newlines))
+			(format *stderr* "~NC" (- (+ prompt-row newlines) last-row) #\newline)
+			(set! prompt-row (- prompt-row newlines)))
+		      (set! cur-row newlines)))))
+
+	      (define (get-previous-line c)         ; get earlier line indexed by numeric arg
+		(let* ((len (length histtop))
+		       (pos (or (string->number cur-line) len)))
+		  (set! pos (min len (max pos 1)))
+		  (set! cur-line (copy (histtop (- pos 1))))
+		  (fixup-new-line)))
+	      
+	      (define (move-forward-in-history c)
+		(set! m-p-pos (min 0 (+ m-p-pos 1)))
+		(when (positive? (length (history m-p-pos)))
+		  (set! cur-line (history m-p-pos))
+		  (fixup-new-line)))
+	      
+	      (define (move-backward-in-history c)
+		(let ((old-index m-p-pos))
+		  (when (and (zero? m-p-pos) 
+			     (> (length cur-line) 0))
+		    (set! (history) cur-line)
+		    (set! m-p-pos -1))
+		  (set! m-p-pos (max (- m-p-pos 1) (- histsize)))
+		  (if (positive? (length (history m-p-pos)))
+		      (begin
+			(set! cur-line (history m-p-pos))
+			(fixup-new-line))
+		      (set! m-p-pos old-index))))
+	      
+	      (define (go-to-start c)
+		(set! cursor-pos 0)
+		'just-cursor)
+	      
+	      (define (go-to-end c)
 		(set! cursor-pos (length cur-line))
-		(let ((newlines (count-newlines cur-line)))
-		  (when (< last-row (+ prompt-row newlines))
-		    (format *stderr* "~NC" (- (+ prompt-row newlines) last-row) #\newline)
-		    (set! prompt-row (- prompt-row newlines)))
-		  (set! cur-row newlines)))
-
-		(define (get-previous-line c)         ; get earlier line indexed by numeric arg
-		  (let* ((len (length histtop))
-			 (pos (or (string->number cur-line) len)))
-		    (set! pos (min len (max pos 1)))
-		    (set! cur-line (copy (histtop (- pos 1))))
-		    (fixup-new-line)))
-
-		(define (move-forward-in-history c)
-		  (set! m-p-pos (min 0 (+ m-p-pos 1)))
-		  (when (positive? (length (history m-p-pos)))
-		    (set! cur-line (history m-p-pos))
-		    (fixup-new-line)))
-
-		(define (move-backward-in-history c)
-		  (let ((old-index m-p-pos))
-		    (when (and (zero? m-p-pos) 
-			       (> (length cur-line) 0))
-		      (set! (history) cur-line)
-		      (set! m-p-pos -1))
-		    (set! m-p-pos (max (- m-p-pos 1) (- histsize)))
-		    (if (positive? (length (history m-p-pos)))
-			(begin
-			  (set! cur-line (history m-p-pos))
-			  (fixup-new-line))
-			(set! m-p-pos old-index))))
+		'just-cursor)
+	      
+	      (define (capitalize c)
+		(let ((len (length cur-line)))
+		  (let loop ((i cursor-pos))
+		    (if (< i len)
+			(if (char-alphabetic? (cur-line i))
+			    (begin
+			      (save-line)
+			      (set! (cur-line i) (char-upcase (cur-line i)))
+			      (set! cursor-pos (word-break i)))
+			    (loop (+ i 1)))))))
+	      
+	      (define (upper-case c)
+		(do ((len (length cur-line))
+		     (i cursor-pos (+ i 1)))
+		    ((or (= i len)
+			 (char-alphabetic? (cur-line i)))
+		     (when (< i len)
+		       (save-line)
+		       (do ((k i (+ k 1)))
+			   ((or (= k len)
+				(not (char-alphabetic? (cur-line k))))
+			    (set! cursor-pos k))
+			 (set! (cur-line k) (char-upcase (cur-line k))))))))
+	      
+	      (define (lower-case c)
+		(do ((len (length cur-line))
+		     (i cursor-pos (+ i 1)))
+		    ((or (= i len)
+			 (char-alphabetic? (cur-line i)))
+		     (when (< i len)
+		       (save-line)
+		       (do ((k i (+ k 1)))
+			   ((or (= k len)
+				(not (char-alphabetic? (cur-line k))))
+			    (set! cursor-pos k))
+			 (set! (cur-line k) (char-downcase (cur-line k))))))))
+	      
+	      (set! (meta-keymap-functions (char->integer #\p)) move-backward-in-history)
+	      (set! (meta-keymap-functions (char->integer #\n)) move-forward-in-history)
+	      (set! (meta-keymap-functions (char->integer #\.)) get-previous-line)
+	      (set! (meta-keymap-functions (char->integer #\<)) go-to-start)
+	      (set! (meta-keymap-functions (char->integer #\>)) go-to-end)
+	      (set! (meta-keymap-functions (char->integer #\c)) capitalize)
+	      (set! (meta-keymap-functions (char->integer #\u)) upper-case)
+	      (set! (meta-keymap-functions (char->integer #\l)) lower-case))
 	    
-		(define (go-to-start c)
-		  (set! cursor-pos 0)
-		  'just-cursor)
+	    ;; -------- emacs --------
+	    (define (emacs-repl)
+	      ;; someday maybe use the language server protocol? (not our own rpc stuff or epc), for json, see json.scm 
+	      ;;   probably will need an argument/function for repl to open the channel or whatever
+	      ;; also this does not resend the entire expression after editing
+	      ;;   and does not notice in-place edits
+	      ;;   can <cr> get entire expr?
+	      (let ((buf (c-pointer->string (calloc 512 1) 512)))
+		(format *stderr* "> ")
+		(do ((b (fgets buf 512 stdin) (fgets buf 512 stdin)))
+		    ((zero? (length b))
+		     (#_exit))
+		  (let ((len (strlen buf)))
+		    (when (positive? len)
+		      (do ((i 0 (+ i 1)))
+			  ((or (not (char-whitespace? (buf i)))
+			       (= i len))
+			   (when (< i len)
+			     (let ((str (substring buf 0 (- (strlen buf) 1))))
+			       ;(format *stderr* "str: ~S~%" str)
+			       (catch #t
+				 (lambda ()
+				   (do ()
+				       ((= (string-length str) 0))
+				     (catch 'string-read-error
+				       (lambda ()
+					 (with-repl-let
+					  (lambda ()
+					    (format *stderr* "~S~%> " (eval-string str (*repl* 'top-level-let)))
+					    (set! str ""))))
+				       (lambda (type info)
+					 (fgets buf 512 stdin)
+					 (set! str (string-append str " " (substring buf 0 (- (strlen buf) 1))))))))
+				 (lambda (type info)
+				   (set! str "")
+				   (apply format *stderr* info)
+				   (format *stderr* "~%> "))))))))))))
 
-		(define (go-to-end c)
-		  (set! cursor-pos (length cur-line))
-		  'just-cursor)
-
-		(define (capitalize c)
-		  (let ((len (length cur-line)))
-		    (let loop ((i cursor-pos))
-		      (if (< i len)
-			  (if (char-alphabetic? (cur-line i))
-			      (begin
-				(save-line)
-				(set! (cur-line i) (char-upcase (cur-line i)))
-				(set! cursor-pos (word-break i)))
-			      (loop (+ i 1)))))))
-		
-		(define (upper-case c)
-		  (do ((len (length cur-line))
-		       (i cursor-pos (+ i 1)))
-		      ((or (= i len)
-			   (char-alphabetic? (cur-line i)))
-		       (when (< i len)
-			 (save-line)
-			 (do ((k i (+ k 1)))
-			     ((or (= k len)
-				  (not (char-alphabetic? (cur-line k))))
-			      (set! cursor-pos k))
-			   (set! (cur-line k) (char-upcase (cur-line k))))))))
-	    
-		(define (lower-case c)
-		  (do ((len (length cur-line))
-		       (i cursor-pos (+ i 1)))
-		      ((or (= i len)
-			   (char-alphabetic? (cur-line i)))
-		       (when (< i len)
-			 (save-line)
-			 (do ((k i (+ k 1)))
-			     ((or (= k len)
-				  (not (char-alphabetic? (cur-line k))))
-			      (set! cursor-pos k))
-			   (set! (cur-line k) (char-downcase (cur-line k))))))))
-		
-		(set! (meta-keymap-functions (char->integer #\p)) move-backward-in-history)
-		(set! (meta-keymap-functions (char->integer #\n)) move-forward-in-history)
-		(set! (meta-keymap-functions (char->integer #\.)) get-previous-line)
-		(set! (meta-keymap-functions (char->integer #\<)) go-to-start)
-		(set! (meta-keymap-functions (char->integer #\>)) go-to-end)
-		(set! (meta-keymap-functions (char->integer #\c)) capitalize)
-		(set! (meta-keymap-functions (char->integer #\u)) upper-case)
-		(set! (meta-keymap-functions (char->integer #\l)) lower-case)
-		)
-	    
-	    ;; -------- terminal setup --------
-	    (define* (run file)
-	      (let ((saved #f)
-		    (tty #t))
+	    ;; -------- rxvt et al --------
+	    (define (terminal-repl file)		
+	      (let ((saved #f))
 		
 		;; we're in libc here, so exit is libc's exit!
 		(define (tty-reset)
-		  (if tty (tcsetattr terminal-fd TCSAFLUSH saved))
+		  (tcsetattr terminal-fd TCSAFLUSH saved)
 		  (if (not (equal? input-fd terminal-fd)) (close input-fd))
 		  (#_exit))
 		
 		(varlet (*repl* 'top-level-let) 
-		  :exit (let ((documentation "(exit) resets the repl tty and exits the repl"))
+		  :exit (let ((+documentation+ "(exit) resets the repl tty and exits the repl"))
 			  (lambda ()
 			    (newline *stderr*)
 			    (tty-reset))))
 		
-		;; check for dumb terminal
-		(if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
-			(string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs shell for example
-		    (let ((buf (c-pointer->string (calloc 512 1) 512)))
-		      (set! tty #f)
-		      (format *stderr* "> ")
-		      (do ((b (fgets buf 512 stdin) (fgets buf 512 stdin)))
-			  ((zero? (length b))
-			   (#_exit))
-			(let ((len (strlen buf)))
-			  (when (positive? len)
-			    (do ((i 0 (+ i 1)))
-				((or (not (char-whitespace? (buf i)))
-				     (= i len))
-				 (when (< i len)
-				   (with-repl-let
-				    (lambda ()
-				      (catch #t
-					(lambda ()
-					  (format *stderr* "~S~%" (eval-string (substring buf 0 (- (strlen buf) 1)) (*repl* 'top-level-let))))
-					(lambda (type info)
-					  (format *stderr* "error: ")
-					  (apply format *stderr* info)
-					  (newline *stderr*)))))
-				   (format *stderr* "> ")))))))))
-		
-		;; not a pipe or a dumb terminal -- hopefully all others accept vt100 codes
+		;; a "normal" terminal -- hopefully it accepts vt100 codes
 		(let ((buf (termios.make)))
 		  (let ((read-size 128))
 		    (set! next-char                                     ; this indirection is needed if user pastes the selection into the repl
@@ -1098,10 +1125,10 @@
 				   (set! chars (read input-fd cc read-size))
 				   (if (= chars 0)
 				       (tty-reset))
-
+				   
 				   (when (> chars (- last-col prompt-length 12))
 				     (let ((str (substring c 0 chars)))
-
+				       
 				       (when (= chars read-size)
 					 ;; concatenate buffers until we get the entire selection
 					 (let reading ((num (read input-fd cc read-size)))
@@ -1136,8 +1163,10 @@
 						    (unless (char=? (str pos) #\tab)
 						      (set! max-cols (max max-cols (- pos start)))
 						      (set! start pos)))
-						  (if (< (+ old-pos 1) (length str))
-						      (set! cur-line (string-append cur-line (substring str (+ old-pos 1)))))
+						  (when (< (+ old-pos 1) (length str))
+						    (set! cur-line (if (zero? old-pos)
+								       (copy str)
+								       (string-append cur-line (substring str (+ old-pos 1))))))
 						  
 						  ;; if the line is too long, the cursor gets confused, so try to reformat long strings
 						  ;;    this still messes up sometimes
@@ -1155,17 +1184,17 @@
 						(display-lines)
 						(return #\newline)))))
 				       ;; now the pasted-in line has inserted newlines, we hope
-
+				       
 				       (set! c str)
 				       (set! cc (string->c-pointer c))
-
+				       
 				       ;; avoid time-consuming redisplays.  We need to use a recursive call on next-char here
 				       ;;   since we might have multi-char commands (embedded #\escape -> meta, etc)
 				       ;; actually, the time is not the repl's fault -- xterm seems to be waiting
 				       ;;   for the window manager or someone to poke it -- if I move the mouse,
 				       ;;   I get immediate output.  I also get immediate output in any case in OSX.
 				       ;;   valgrind and ps say we're not computing, we're just sitting there.
-
+				       
 				       (catch #t
 					 (lambda ()
 					   (do ((ch (next-char) (next-char)))
@@ -1177,7 +1206,7 @@
 					 
 					 (lambda (type info)
 					   (set! chars 0)
-					   (move-cursor prompt-row prompt-col)
+					   ;(move-cursor prompt-row prompt-col)
 					   (format *stderr* "internal error: ")
 					   (apply format *stderr* info)
 					   (format *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
@@ -1205,8 +1234,6 @@
 		  (when (negative? (tcsetattr terminal-fd TCSAFLUSH buf))
 		    (tty-reset))
 		  
-		  
-		  ;; -------- the repl --------
 		  (display-prompt)
 		  (cursor-bounds)
 		  ;; (debug-help)
@@ -1228,18 +1255,33 @@
 			  (help chr)))
 		      
 		      (lambda (type info)
-			(move-cursor prompt-row prompt-col)
+			;(move-cursor prompt-row prompt-col)
 			(format *stderr* "internal error: ")
 			(apply format *stderr* info)
 			(format *stderr* "~%line ~A: ~A~%" ((owlet) 'error-line) ((owlet) 'error-code))
 			(set! chars 0)
 			(new-prompt)))))))
 	    
+	    (define* (run file)
+	      ;; check for dumb terminal
+	      (if (or (zero? (isatty terminal-fd))        ; not a terminal -- input from pipe probably
+		      (string=? (getenv "TERM") "dumb"))  ; no vt100 codes -- emacs subjob for example
+		  (emacs-repl)                            ; TODO: restore support for the pipe case
+		  (terminal-repl file)))
+
 	    (curlet))))))
       
-      (define (save-repl) 
-	(call-with-output-file "save.repl" 
+      (define* (save-repl (file "save.repl"))
+	;; ((*repl* 'save-repl))
+	(call-with-output-file file
 	  (lambda (p) 
+	    (format p ";;; save-repl ~A~%~%" 
+		    (with-let (sublet *libc*)
+		      (let ((timestr (make-string 128))) 
+			(let ((len (strftime timestr 128 "%a %d-%b-%Y %H:%M:%S %Z"
+					     (localtime 
+					      (time.make (time (c-pointer 0 'time_t*)))))))
+			  (substring timestr 0 len)))))
 	    (format p "(for-each~%~NC~
                          (lambda (f)~%~NC~
                            (if (not (provided? f))~%~NC~
@@ -1250,13 +1292,15 @@
 	    (format p "(with-let (*repl* 'repl-let)~%")
 	    (((*repl* 'repl-let) 'write-history) p 2)
 	    (format p ")~%~%")
-	    (format p "~W" (*repl* 'top-level-let)))))
+	    (for-each (lambda (var)
+			(unless (eq? (car var) 'exit)
+			  (format p "(define ~S ~W)~%" (car var) (cdr var))))
+		      (*repl* 'top-level-let)))))
       
       (define (restore-repl) 
 	(set! (*repl* 'top-level-let) (load "save.repl")))  
       ;; I think this could be a merge rather than a reset by using (with-let top-level-let (load ...))
       
-      
       (set! keymap (repl-let 'keymap))
       (set! history (repl-let 'history))
       (set! history-size (repl-let 'history-size))
@@ -1293,7 +1337,7 @@
 		(let ((timestr (make-string 128))) 
 		  (let ((len (strftime timestr 128 "%a %d-%b-%Y %H:%M:%S %Z"
 				       (localtime 
-					(time.make (time (c-pointer 0)))))))
+					(time.make (time (c-pointer 0 'time_t*)))))))
 		    (substring timestr 0 len)))))))
     (openlet (inlet 'object->string pd 'let-ref-fallback pd))))
 |#
@@ -1335,85 +1379,88 @@
 	    (* 0.000001 (- (cadr end) (cadr ,start))))))))
 
 
-(define* (apropos name (e (*repl* 'top-level-let)))
-  
-  (define (levenshtein s1 s2)
-    (let ((L1 (length s1))
-	  (L2 (length s2)))
-      (cond ((zero? L1) L2)
-	    ((zero? L2) L1)
-	    (else (let ((distance (make-vector (list (+ L2 1) (+ L1 1)) 0)))
-		    (do ((i 0 (+ i 1)))
-			((> i L1))
-		      (set! (distance 0 i) i))
-		    (do ((i 0 (+ i 1)))
-			((> i L2))
-		      (set! (distance i 0) i))
-		    (do ((i 1 (+ i 1)))
-			((> i L2))
-		      (do ((j 1 (+ j 1)))
-			  ((> j L1))
-			(let ((c1 (+ (distance i (- j 1)) 1))
-			      (c2 (+ (distance (- i 1) j) 1))
-			      (c3 (+ (distance (- i 1) (- j 1)) 
-				     (if (char=? (s2 (- i 1)) (s1 (- j 1))) 0 1))))
-			  (set! (distance i j) (min c1 c2 c3)))))
-		    (distance L2 L1))))))
-
-  (define* (make-full-let-iterator lt (stop (rootlet))) ; walk the entire let chain
-    (if (eq? stop lt)
-	(make-iterator lt)
-	(letrec ((iterloop 
-		  (let ((iter (make-iterator lt))
-			(iterator? #t))
-		    (lambda ()
-		      (let ((result (iter)))
-			(if (and (eof-object? result)
-				 (iterator-at-end? iter)
-				 (not (eq? stop (iterator-sequence iter))))
-			    (begin 
-			      (set! iter (make-iterator (outlet (iterator-sequence iter))))
-			      (iterloop))
-			    result))))))
-	    (make-iterator iterloop))))
-  
-  (let ((ap-name (if (string? name) name 
-		     (if (symbol? name) (symbol->string name)
-			 (error 'wrong-type-arg "apropos argument 1 should be a string or a symbol"))))
-	(ap-env (if (let? e) e 
-		    (error 'wrong-type-arg "apropos argument 2 should be an environment"))))
-    (let ((strs ())
-	  (min2 (floor (log (length ap-name) 2)))
-	  (have-orange (string=? ((*libc* 'getenv) "TERM") "xterm-256color")))
-      (for-each
-       (lambda (binding)
-	 (if (pair? binding)
-	     (let ((symbol-name (symbol->string (car binding))))
-	       (if (string-position ap-name symbol-name)
-		   (set! strs (cons (cons binding 0) strs))
-		   (let ((distance (levenshtein ap-name symbol-name)))
-		     (if (< distance min2)
-			 (set! strs (cons (cons binding distance) strs))))))))
-       (make-full-let-iterator ap-env))
-
-      (if (not (pair? strs))
-	  'no-match
-	  (begin
-	    (for-each (lambda (b)
-			(format *stderr*
-				(if (zero? (cdr b)) "~C[1m~A~C[0m: ~S~%"                           ; black if exact match somewhere
-				    (if (or (< (cdr b) 2) (not have-orange)) "~C[31m~A~C[0m: ~S~%" ; red for near miss
-					"~C[38;5;208m~A~C[0m: ~S~%"))                              ; orange for less likely choices
-				#\escape (caar b) #\escape
-				(if (procedure? (cdar b))
-				    (let ((doc (procedure-documentation (cdar b)))) ; returns "" if no doc
-				      (if (positive? (length doc))
-					  doc
-					  'procedure))
-				    (cdar b))))
-		      (sort! strs (lambda (a b)
-				    (string<? (symbol->string (caar a)) (symbol->string (caar b))))))
-	    '----)))))
+(define apropos 
+  (let ((levenshtein 
+	 (lambda (s1 s2)
+	   (let ((L1 (length s1))
+		 (L2 (length s2)))
+	     (cond ((zero? L1) L2)
+		   ((zero? L2) L1)
+		   (else (let ((distance (make-vector (list (+ L2 1) (+ L1 1)) 0)))
+			   (do ((i 0 (+ i 1)))
+			       ((> i L1))
+			     (set! (distance 0 i) i))
+			   (do ((i 0 (+ i 1)))
+			       ((> i L2))
+			     (set! (distance i 0) i))
+			   (do ((i 1 (+ i 1)))
+			       ((> i L2))
+			     (do ((j 1 (+ j 1)))
+				 ((> j L1))
+			       (let ((c1 (+ (distance i (- j 1)) 1))
+				     (c2 (+ (distance (- i 1) j) 1))
+				     (c3 (if (char=? (s2 (- i 1)) (s1 (- j 1)))
+					     (distance (- i 1) (- j 1))
+					     (+ (distance (- i 1) (- j 1)) 1))))
+				 (set! (distance i j) (min c1 c2 c3)))))
+			   (distance L2 L1)))))))
+	
+	(make-full-let-iterator             ; walk the entire let chain
+	 (lambda* (lt (stop (rootlet))) 
+	   (if (eq? stop lt)
+	       (make-iterator lt)
+	       (letrec ((iterloop 
+			 (let ((iter (make-iterator lt))
+			       (+iterator+ #t))
+			   (lambda ()
+			     (let ((result (iter)))
+			       (if (and (eof-object? result)
+					(iterator-at-end? iter)
+					(not (eq? stop (iterator-sequence iter))))
+				   (begin 
+				     (set! iter (make-iterator (outlet (iterator-sequence iter))))
+				     (iterloop))
+				   result))))))
+		 (make-iterator iterloop))))))
+    
+    (lambda* (name (e (*repl* 'top-level-let)))
+      (let ((ap-name (if (string? name) name 
+			 (if (symbol? name) (symbol->string name)
+			     (error 'wrong-type-arg "apropos argument 1 should be a string or a symbol"))))
+	    (ap-env (if (let? e) e 
+			(error 'wrong-type-arg "apropos argument 2 should be an environment"))))
+	(let ((strs ())
+	      (min2 (floor (log (length ap-name) 2)))
+	      (have-orange (string=? ((*libc* 'getenv) "TERM") "xterm-256color")))
+	  (for-each
+	   (lambda (binding)
+	     (if (pair? binding)
+		 (let ((symbol-name (symbol->string (car binding))))
+		   (if (string-position ap-name symbol-name)
+		       (set! strs (cons (cons binding 0) strs))
+		       (let ((distance (levenshtein ap-name symbol-name)))
+			 (if (< distance min2)
+			     (set! strs (cons (cons binding distance) strs))))))))
+	   (make-full-let-iterator ap-env))
+	  
+	  (if (not (pair? strs))
+	      'no-match
+	      (begin
+		(for-each (lambda (b)
+			    (format *stderr*
+				    (if (zero? (cdr b)) "~C[1m~A~C[0m: ~S~%"                           ; black if exact match somewhere
+					(if (or (< (cdr b) 2) (not have-orange)) "~C[31m~A~C[0m: ~S~%" ; red for near miss
+					    "~C[38;5;208m~A~C[0m: ~S~%"))                              ; orange for less likely choices
+				    #\escape (caar b) #\escape
+				    (if (procedure? (cdar b))
+					(let ((doc (documentation (cdar b)))) ; returns "" if no doc
+					  (if (positive? (length doc))
+					      doc
+					      'procedure))
+					(cdar b))))
+			  (sort! strs (lambda (a b)
+					(string<? (symbol->string (caar a)) (symbol->string (caar b))))))
+		'----)))))))
 
 
 
@@ -1473,13 +1520,13 @@
 
 ;;; to display a variable's value as s7 runs using the repl help window:
 ;;;    (define xyz 1) ; some variable...
-;;;    (set! (symbol-access 'xyz) (lambda (sym val) (set! (*repl* 'helpers) (list (lambda (c) (format #f "xyz: ~S" val)))) val))
+;;;    (set! (setter 'xyz) (lambda (sym val) (set! (*repl* 'helpers) (list (lambda (c) (format #f "xyz: ~S" val)))) val))
 
 
 ;;; --------------------------------------------------------------------------------
 #|
 to work in a particular environment:
-    (set! (*repl* 'top-level-let) (sublet (rootlet) :** #f)) ; or any other like *libc*
+    (set! (*repl* 'top-level-let) (sublet (rootlet))) ; or any other like *libc*
 now (define g 43) puts g in the new top-level-let, so ((rootlet) 'g) -> #<undefined>, and ((*repl* 'top-level-let) 'g) -> 43 (= g in repl of course)
 to start with a fresh top-level, just set top-level-let to (sublet (rootlet)) again.
 
